struct bad_alloc {}; // Remove if your compiler has a built-in bad_alloc
struct stack_underflow {};
struct pc_error {}; // used for post condition violations.

class intstack
{
public:
  intstack() throw ();
  // Preconditions: -

  intstack(const intstack& is) throw (bad_alloc);
  // Preconditions: -

  intstack& operator=(const intstack& is) throw (bad_alloc);
  // Preconditions: -

  ~intstack() throw ();
    // Preconditions: -
    // Postconditions:
    //   The memory allocated by the object is deallocated

  unsigned nrOfElements() const throw (pc_error);
    // Preconditions: -
    // Postconditions:
    //   nrOfElements() == 0 || top() == old top()

  void push(int anInt) throw (bad_alloc, pc_error);
    // Preconditions: -
    // Postconditions:
    //   nrOfElements() = 1 + old nrOfElements()
    //   top() == anInt
    // Behaviour on exception:
    //   Stack unchanged.

  int top(void) const throw (stack_underflow, pc_error);
    // Preconditions:
    //   nrOfElements() > 0
    // Postconditions:
    //   nrOfElements() ==  old nrOfElements()
    // Behaviour on exception:
    //   Stack unchanged.

  int& top(void) throw (stack_underflow, pc_error);
    // Preconditions:
    //   nrOfElements() > 0
    // Postconditions:
    //   nrOfElements() ==  old nrOfElements()
    // Behaviour on exception:
    //   Stack unchanged.

  void pop(void) throw (stack_underflow, pc_error);
    // Preconditions:
    //   nrOfElements() > 0
    // Postconditions:
    //   nrOfElements() + 1 == old nrOfElements()
    // Behaviour on exception:
    //   Stack unchanged.
private:
  struct stack_element
  {
    stack_element(int aValue, stack_element* p) throw ()
      : value(aValue), pNext(p) {};
    int value;
    stack_element* pNext;
  };
  stack_element* pTop;
  unsigned elements;

  // helper functions for copy constructor, copy
  // assignment and destructor.
  stack_element* copy(void) const throw (bad_alloc);
  void deallocateAll(void) throw();
};


intstack::intstack() throw ()
  : pTop(0),
    elements(0)
{
    // Preconditions: -
}

intstack::intstack(const intstack& i) throw (bad_alloc)
  : pTop(i.copy()),
    elements(i.elements)
{
  // Preconditions: -
  // Postconditions:
}

intstack& intstack::operator=(const intstack& i)
 throw (bad_alloc)
{
  if (this != &i)
  {
    stack_element* p = i.copy();
    deallocateAll();
    pTop=p;
    elements=i.elements;
  }
  return *this;
}

intstack::~intstack() throw ()
{
  // Preconditions: -
  // Postconditions:
  //   The memory allocated by the object is deallocated

  deallocateAll();
}

void intstack::deallocateAll() throw ()
{
  while (pTop != 0)
  {
    stack_element* p = pTop->pNext;
    delete pTop; // guaranteed not to throw.
    pTop = p;
  }
}

intstack::stack_element* intstack::copy(void) const throw (bad_alloc)
{
  stack_element* pFirst = 0; // used in catch block.
  try {    
    stack_element* p = pTop;
    if (p != 0)
    { // take care of first element here.
      stack_element* pPrevious = 0;
      pFirst = new stack_element(p->value,0); // Cannot throw anything
      if (pFirst == 0)                        // but bad_alloc
	throw bad_alloc();
      pPrevious = pFirst;
      while ((p = p->pNext) != 0)
      {
	pPrevious->pNext = new stack_element(p->value,0); // cannot throw
	pPrevious = pPrevious->pNext; // anything but bad_alloc
	if (pPrevious == 0)
	  throw bad_alloc();
      }
    }
    return pFirst;
  }
  catch (...) // If anything went wrong, deallocate it all
  {           // and rethrow!
    while (pFirst != 0)
    {
      stack_element* pTmp = pFirst->pNext;
      delete pFirst; // guaranteed not to throw.
      pFirst = pTmp;
    }
    throw;
  }
}


unsigned intstack::nrOfElements() const throw (pc_error)
{
  // Preconditions: -
  return elements; 


  // Postconditions:
  //   nrOfElements() == 0 || top() == old top()
 
  // no need to check anything with this
  // implementation as it's trivially
  // obvious that nothing will change.
}

void intstack::push(int anInt) throw (bad_alloc, pc_error)
{
  // Preconditions: -

  unsigned old_nrOfElements = nrOfElements();
  stack_element* pOld = pTop;
  
  stack_element* pTmp = new stack_element(anInt, pTop);
  if (pTmp == 0)
    throw bad_alloc();

  // the above either throws or succeeds. If it throws,
  // the memory is deallocated and we're leaving the
  // function with the exception (before assigning to
  // pTop, so the stack remains unchanged.)

  try {
    pTop = pTmp;
    ++elements;
  
    // Postconditions:
    //   nrOfElements() == 1 + old nrOfElements()
    //   top() == anInt
    
    if (nrOfElements() != 1 + old_nrOfElements
     || top() != anInt)
    {
      throw pc_error();
    }
  } 
  catch (...) {
    // Behaviour on exception:
    //   Stack unchanged.

    delete pTop; // get rid of the new top element
    
    pTop = pOld;
    elements = old_nrOfElements;
    throw;
  }
}

int intstack::top(void) const
  throw (stack_underflow, pc_error)
{
  // Preconditions:
  //   nrOfElements() > 0
  if (nrOfElements() == 0 || pTop == 0)
  {
    throw stack_underflow();
  }
  return pTop->value;
  // Postconditions:
  //   nrOfElements() ==  old nrOfElements()
  // No need to check with this implementation!
}

int& intstack::top(void) 
   throw (stack_underflow, pc_error)
{
  // Preconditions:
  //   nrOfElements() > 0
  if (nrOfElements() == 0 || pTop == 0)
  {
    throw stack_underflow();
  }
  return pTop->value;
  // Postconditions:
  //   nrOfElements() ==  old nrOfElements()
  // No need to check with this implementation!
}

void intstack::pop(void) throw (stack_underflow, pc_error)
{
  // Preconditions:
  //   nrOfElements() > 0
  
  if (nrOfElements() == 0 || pTop == 0)
  {
    throw stack_underflow();
  }
  
  unsigned old_elements = nrOfElements();
  stack_element* pOld = pTop;

  try {    
    pTop = pTop->pNext;
    --elements;

    // Postconditions:
    //   nrOfElements() + 1 == old nrOfElements()
  
    if (nrOfElements() + 1 != old_elements)
    {
      // Behaviour on exception:
      //   Stack unchanged.

      throw pc_error();
    }
  }
  catch (...) {
    elements = old_elements;
    pTop = pOld;
    throw;
  }  
  delete pOld; // guaranteed not to throw.
}

#include <iostream.h>

template <class T>
void print(const T& t)
{
  cout << "t=" << t << endl;
}

int main(int argc, char* argv[])
{
  intstack is;
  print(is);
  return 0;
}
